﻿#region Copyright
//
// Copyright (C) 2010-2011 by Autodesk, Inc.
//
// Permission to use, copy, modify, and distribute this software in
// object code form for any purpose and without fee is hereby granted,
// provided that the above copyright notice appears in all copies and
// that both that copyright notice and the limited warranty and
// restricted rights notice below appear in all supporting
// documentation.
//
// AUTODESK PROVIDES THIS PROGRAM "AS IS" AND WITH ALL FAULTS.
// AUTODESK SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTY OF
// MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.  AUTODESK, INC.
// DOES NOT WARRANT THAT THE OPERATION OF THE PROGRAM WILL BE
// UNINTERRUPTED OR ERROR FREE.
//
// Use, duplication, or disclosure by the U.S. Government is subject to
// restrictions set forth in FAR 52.227-19 (Commercial Computer
// Software - Restricted Rights) and DFAR 252.227-7013(c)(1)(ii)
// (Rights in Technical Data and Computer Software), as applicable.
//
// 
#endregion // Copyright

#region Imports

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Autodesk.Revit.DB;
using Autodesk.Revit.UI;
using Autodesk.Revit.Attributes;
using Autodesk.Revit.UI.Selection; // for selection
using Autodesk.Revit.ApplicationServices; // Application class
using Autodesk.Revit.DB.Analysis; 

#endregion

namespace DevDay2010
{
  /// <summary>
  /// Geometry sample 2: creation of a revolved solid and using it from the AVF.
  /// </summary>
  [Transaction( TransactionMode.Manual )]
  public class Cmd_GeometryApiAvf : IExternalCommand
  {
    UIApplication m_rvtUIApp;
    UIDocument m_rvtUIDoc;
    Application m_rvtApp;
    Document m_rvtDoc;

    public Result Execute(
      ExternalCommandData commandData,
      ref string message,
      ElementSet elements )
    {
      //  get the access to the top most objects. 
      //  Notice that we have UI abd DB application and Document. 
      //  (we are not using them all here. This is to show two versions.) 
      m_rvtUIApp = commandData.Application;
      m_rvtUIDoc = m_rvtUIApp.ActiveUIDocument;
      m_rvtApp = m_rvtUIApp.Application;
      m_rvtDoc = m_rvtUIDoc.Document;

      //  this is the main function. 
      myCommand();

      return Result.Succeeded;
    }

    public void myCommand()
    {
      Transaction trans = new Transaction( m_rvtUIDoc.Document );
      trans.Start( "TestAFV" );

      createAVFDisplayStyle();

      XYZ pt1 = new XYZ( 2, 2, 0 );
      XYZ pt2 = new XYZ( 10, 10, 0 );
      CreateRevolvedRectangle( pt1, pt2, 10 );
      trans.Commit();
    }

    private void createAVFDisplayStyle()
    {
      string myDisplayStyleName = "My Colored Surface";
      AnalysisDisplayStyle analysisDisplayStyle = null;

      // Look for an existing analysis display style with a specific name
      FilteredElementCollector collector1 = new FilteredElementCollector( m_rvtDoc );
      ICollection<Element> collection = collector1.OfClass( typeof( AnalysisDisplayStyle ) ).ToElements();
      var displayStyle = from element in collection
                         where element.Name == myDisplayStyleName
                         select element;
      //  if we have a style already, use it. 
      if( displayStyle.Count() > 0 )
      {
        return;
      }

      // If display style does not already exist in the document, create it

      AnalysisDisplayColoredSurfaceSettings coloredSurfaceSettings = new AnalysisDisplayColoredSurfaceSettings();
      AnalysisDisplayColorSettings colorSettings = new AnalysisDisplayColorSettings();
      AnalysisDisplayLegendSettings legendSettings = new AnalysisDisplayLegendSettings();
      //AnalysisDisplayStyle analysisDisplayStyle = AnalysisDisplayStyle.CreateAnalysisDisplayStyle(m_rvtDoc, "Colored Surface", coloredSurfaceSettings, colorSettings, legendSettings);
      analysisDisplayStyle = AnalysisDisplayStyle.CreateAnalysisDisplayStyle( m_rvtDoc, myDisplayStyleName, coloredSurfaceSettings, colorSettings, legendSettings );

      m_rvtDoc.ActiveView.AnalysisDisplayStyleId = analysisDisplayStyle.Id;
    }

    private void CreateRevolvedRectangle( XYZ firstCorner, XYZ secondCorner, double value )
    {
      try
      {
        List<CurveLoop> loops = new List<CurveLoop>();

        CurveLoop loop1 = new CurveLoop();
        Autodesk.Revit.Creation.Application appCreate = m_rvtApp.Create;

        XYZ corner2 = new XYZ( secondCorner.X, firstCorner.Y, 0 );
        XYZ corner4 = new XYZ( firstCorner.X, secondCorner.Y, 0 );
        loop1.Append( appCreate.NewLineBound( firstCorner, corner2 ) );
        loop1.Append( appCreate.NewLineBound( corner2, secondCorner ) );
        loop1.Append( appCreate.NewLineBound( secondCorner, corner4 ) );
        loop1.Append( appCreate.NewLineBound( corner4, firstCorner ) );

        //loop1.Append(appCreate.NewArc(new XYZ(0, 10, 0), 10, 0, -Math.PI/2, XYZ.BasisX, XYZ.BasisY));
        loops.Add( loop1 );

        Frame frame = new Frame();
        frame.BasisX = XYZ.BasisX;
        frame.BasisY = XYZ.BasisZ;
        frame.BasisZ = -XYZ.BasisY;

        Solid revolution = GeometryCreationUtilities.CreateRevolvedGeometry( frame, loops, 0, Math.PI * 2 );
        m_rvtDoc.Regenerate();
        PaintSolid( revolution, value );
      }
      catch( Exception ex )
      {
        TaskDialog.Show( "Exception", ex.Message );
      }
    }

    private static int schemaId = -1;

    private void PaintSolid( Solid s, double value )
    {
      try
      {
        Application app = m_rvtApp;

        SpatialFieldManager sfm = SpatialFieldManager.GetSpatialFieldManager( 
          m_rvtDoc.ActiveView );

        if( sfm == null ) sfm = SpatialFieldManager.CreateSpatialFieldManager( 
          m_rvtDoc.ActiveView, 1 );

        if( schemaId != -1 )
        {
          IList<int> results = sfm.GetRegisteredResults();

          if( !results.Contains( schemaId ) )
          {
            schemaId = -1;
          }
        }

        if( schemaId == -1 )
        {
          AnalysisResultSchema resultSchema1 
            = new AnalysisResultSchema( "PaintedSolid", "Description" );

          schemaId = sfm.RegisterResult( resultSchema1 );
        }

        FaceArray faces = s.Faces;
        Transform trf = Transform.Identity;

        foreach( Face face in faces )
        {
          int idx = sfm.AddSpatialFieldPrimitive( face, trf );

          IList<UV> uvPts = new List<UV>();
          List<double> doubleList = new List<double>();
          IList<ValueAtPoint> valList = new List<ValueAtPoint>();
          BoundingBoxUV bb = face.GetBoundingBox();
          uvPts.Add( bb.Min );
          doubleList.Add( value );
          valList.Add( new ValueAtPoint( doubleList ) );
          FieldDomainPointsByUV pnts = new FieldDomainPointsByUV( uvPts );
          FieldValues vals = new FieldValues( valList );

          sfm.UpdateSpatialFieldPrimitive( idx, pnts, vals, schemaId );
        }
      }
      catch( Exception ex )
      {
        TaskDialog.Show( "test", ex.Message );
      }
    }
  }
}
