﻿#region Copyright
//
// Copyright (C) 2010-2011 by Autodesk, Inc.
//
// Permission to use, copy, modify, and distribute this software in
// object code form for any purpose and without fee is hereby granted,
// provided that the above copyright notice appears in all copies and
// that both that copyright notice and the limited warranty and
// restricted rights notice below appear in all supporting
// documentation.
//
// AUTODESK PROVIDES THIS PROGRAM "AS IS" AND WITH ALL FAULTS.
// AUTODESK SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTY OF
// MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.  AUTODESK, INC.
// DOES NOT WARRANT THAT THE OPERATION OF THE PROGRAM WILL BE
// UNINTERRUPTED OR ERROR FREE.
//
// Use, duplication, or disclosure by the U.S. Government is subject to
// restrictions set forth in FAR 52.227-19 (Commercial Computer
// Software - Restricted Rights) and DFAR 252.227-7013(c)(1)(ii)
// (Rights in Technical Data and Computer Software), as applicable.
//
// 
#endregion // Copyright

#region Imports

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Autodesk.Revit.DB;
using Autodesk.Revit.UI;
using Autodesk.Revit.Attributes;
using Autodesk.Revit.UI.Selection; // for selection
using Autodesk.Revit.ApplicationServices; // Application class

#endregion

namespace DevDay2010
{

  /// <summary>
  /// Geometry sample 1: ElementIntersectsSolidFilter
  /// This example uses ElementIntersectsSolidFilter applied 
  /// against a cylinder created at the end of a wall to find 
  /// other walls in close proximity to the wall endpoints,
  /// whether or not they are actually joined.
  /// </summary>
  [Transaction( TransactionMode.Automatic )]
  public class Cmd_GeometryFilter : IExternalCommand
  {
    UIApplication m_rvtUIApp;
    UIDocument m_rvtUIDoc;
    Application m_rvtApp;
    Document m_rvtDoc;

    public Result Execute( 
      ExternalCommandData commandData, 
      ref string message, 
      ElementSet elements )
    {
      //  get the access to the top most objects. 
      //  Notice that we have UI abd DB application and Document. 
      //  (we are not using them all here. This is to show two versions.) 
      m_rvtUIApp = commandData.Application;
      m_rvtUIDoc = m_rvtUIApp.ActiveUIDocument;
      m_rvtApp = m_rvtUIApp.Application;
      m_rvtDoc = m_rvtUIDoc.Document;

      //  this is the main function. 
      testWallIntersect();

      return Result.Succeeded;
    }

    public void testWallIntersect()
    {
      Selection sel = m_rvtUIDoc.Selection;
      Reference ref1 = sel.PickObject( ObjectType.Element, "Pick a wall" );

      Element elem = m_rvtUIDoc.Document.GetElement( ref1 );
      Wall wall = elem as Wall;

      int nend = 0;
      IEnumerable<Element> a = FindWallJoinsAtEndUsingProximity( wall, nend );
      string s = "Elements returned: ";
      foreach( Element e in a )
      {
        s += "\r\n" + e.Category.Name + " " + e.Id.IntegerValue.ToString();
      }
      TaskDialog.Show( "Proximity Results", s );
    }

    public IEnumerable<Element> FindWallJoinsAtEndUsingProximity( 
      Wall wall, 
      int end )
    {
      // Get properties of wall at the end point
      LocationCurve wallCurve = wall.Location as LocationCurve;
      XYZ endPoint = wallCurve.Curve.get_EndPoint( end );
      double height = wall.get_Parameter( BuiltInParameter.WALL_USER_HEIGHT_PARAM ).AsDouble();
      double elevation = wall.Level.Elevation;

      // Build cylinder centered at wall end point, extending 3’ in diameter
      CurveLoop cylinderLoop = new CurveLoop();
      XYZ arcCenter = new XYZ( endPoint.X, endPoint.Y, elevation );
      Application application = wall.Document.Application;

      Arc firstArc = application.Create.NewArc( 
        arcCenter, 1.5, 0, Math.PI, XYZ.BasisX, XYZ.BasisY );

      Arc secondArc = application.Create.NewArc( 
        arcCenter, 1.5, Math.PI, 2 * Math.PI, XYZ.BasisX, XYZ.BasisY );

      cylinderLoop.Append( firstArc );
      cylinderLoop.Append( secondArc );

      List<CurveLoop> singleLoop = new List<CurveLoop>();
      singleLoop.Add( cylinderLoop );

      Solid proximityCylinder 
        = GeometryCreationUtilities.CreateExtrusionGeometry( 
          singleLoop, XYZ.BasisZ, height );

      // Filter walls
      FilteredElementCollector proximityCollector 
        = new FilteredElementCollector( wall.Document );

      proximityCollector.OfClass( typeof( Wall ) );

      // Exclude the wall itself
      List<ElementId> toExclude = new List<ElementId>();
      toExclude.Add( wall.Id );
      proximityCollector.Excluding( toExclude );

      // Filter only elements intersecting our target cylinder
      proximityCollector.WherePasses( 
        new ElementIntersectsSolidFilter( 
          proximityCylinder ) );

      // Return matches
      return proximityCollector;
    }
  }
}
