﻿#region Copyright
//
// Copyright (C) 2010-2011 by Autodesk, Inc.
//
// Permission to use, copy, modify, and distribute this software in
// object code form for any purpose and without fee is hereby granted,
// provided that the above copyright notice appears in all copies and
// that both that copyright notice and the limited warranty and
// restricted rights notice below appear in all supporting
// documentation.
//
// AUTODESK PROVIDES THIS PROGRAM "AS IS" AND WITH ALL FAULTS.
// AUTODESK SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTY OF
// MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.  AUTODESK, INC.
// DOES NOT WARRANT THAT THE OPERATION OF THE PROGRAM WILL BE
// UNINTERRUPTED OR ERROR FREE.
//
// Use, duplication, or disclosure by the U.S. Government is subject to
// restrictions set forth in FAR 52.227-19 (Commercial Computer
// Software - Restricted Rights) and DFAR 252.227-7013(c)(1)(ii)
// (Rights in Technical Data and Computer Software), as applicable.
//
// 
#endregion // Copyright

#region Namespaces
using System;
using System.Diagnostics;
using System.Collections.Generic;
using Autodesk.Revit.ApplicationServices;
using Autodesk.Revit.Attributes;
using Autodesk.Revit.DB;
using Autodesk.Revit.DB.ExtensibleStorage;
using Autodesk.Revit.UI;
using Autodesk.Revit.UI.Selection;
#endregion

namespace ExtensibleStorage
{
  /// <summary>
  /// This lab demonstrates the new extensible storage functionality. 
  /// In this example, we store the location of a wall socket into 
  /// extensible data stored on the wall.
  /// Please also look at the ExtensibleStorageManager Revit SDK sample.
  /// </summary>
  [Transaction( TransactionMode.Manual )]
  class ExtensibleStorage : IExternalCommand
  {
    /// <summary>
    /// The schema specific GUID. 
    /// </summary>
    Guid _guid = new Guid( "87aaad89-6f1b-45e1-9397-2985e1560a02" );

    /// <summary>
    /// Allow only walls to be selected.
    /// </summary>
    class WallSelectionFilter : ISelectionFilter
    {
      public bool AllowElement( Element e )
      {
        return e is Wall;
      }

      public bool AllowReference( Reference r, XYZ p )
      {
        return true;
      }
    }

    public Result Execute(
      ExternalCommandData commandData,
      ref string message,
      ElementSet elements )
    {
      UIDocument uidoc = commandData.Application.ActiveUIDocument;
      Document doc = uidoc.Document;

      // create transaction for working with schema

      Transaction trans = new Transaction( doc, "Extensible Storage" );
      trans.Start();

      // select a wall element

      Wall wall = null;

      try
      {
        Reference r = uidoc.Selection.PickObject( ObjectType.Element,
          new WallSelectionFilter() );

        wall = doc.GetElement( r ) as Wall;
      }
      catch( OperationCanceledException )
      {
        message = "Nothing selected; please select a wall to attach extensible data to.";
        return Result.Failed;
      }

      Debug.Assert( null != wall, "expected a wall to be selected" );

      if( null == wall )
      {
        message = "Please select a wall to attach extensible data to.";
        return Result.Failed;
      }

      // create a schema builder

      SchemaBuilder builder = new SchemaBuilder( _guid );

      // set read and write access levels

      builder.SetReadAccessLevel( AccessLevel.Public );
      builder.SetWriteAccessLevel( AccessLevel.Public );

      // Note: if this was set as vendor or application access, 
      // we would have been additionally required to use SetVendorId

      // set name to this schema builder

      builder.SetSchemaName( "WallSocketLocation" );
      builder.SetDocumentation( "Data store for socket related info in a wall" );

      // create field1

      FieldBuilder fieldBuilder1 = builder.AddSimpleField( "SocketLocation", typeof( XYZ ) );

      // set unit type

      fieldBuilder1.SetUnitType( UnitType.UT_Length );

      // add documentation (optional)

      // create field2

      FieldBuilder fieldBuilder2 = builder.AddSimpleField( "SocketNumber", typeof( string ) );

      //fieldBuilder2.SetUnitType(UnitType.UT_Custom);

      // register the schema object

      Schema schema = builder.Finish();

      // create an entity (object) for this schema (class)

      Entity ent = new Entity( schema );
      Field socketLocation = schema.GetField( "SocketLocation" );
      ent.Set<XYZ>( socketLocation, new XYZ( 2, 0, 0 ), DisplayUnitType.DUT_METERS );

      Field socketNumber = schema.GetField( "SocketNumber" );
      ent.Set<string>( socketNumber, "200" );

      wall.SetEntity( ent );

      // Now create another entity (object) for this schema (class)

      Entity ent2 = new Entity( schema );
      Field socketNumber1 = schema.GetField( "SocketNumber" );
      ent2.Set<String>( socketNumber1, "400" );
      wall.SetEntity( ent2 );

      // Note: this will replace the previous entity on the wall 

      // list all schemas in the document

      string s = string.Empty;
      IList<Schema> schemas = Schema.ListSchemas();
      foreach( Schema sch in schemas )
      {
        s += "\r\nSchema Name: " + sch.SchemaName;
      }
      TaskDialog.Show( "Schema details", s );

      // list all Fields for our schema

      s = string.Empty;
      Schema ourSchema = Schema.Lookup( _guid );
      IList<Field> fields = ourSchema.ListFields();
      foreach( Field fld in fields )
      {
        s += "\r\nField Name: " + fld.FieldName;
      }
      TaskDialog.Show( "Field details", s );

      // extract the value for the field we created

      Entity wallSchemaEnt = wall.GetEntity( Schema.Lookup( _guid ) );

      XYZ wallSocketPos = wallSchemaEnt.Get<XYZ>(
        Schema.Lookup( _guid ).GetField( "SocketLocation" ),
        DisplayUnitType.DUT_METERS );

      s = "SocketLocation: " + PointString( wallSocketPos );

      string wallSocketNumber = wallSchemaEnt.Get<String>(
        Schema.Lookup( _guid ).GetField( "SocketNumber" ) );

      s += "\r\nSocketNumber: " + wallSocketNumber;

      TaskDialog.Show( "Field values", s );

      trans.Commit();

      return Result.Succeeded;
    }

    public static string PointString(XYZ p)
    {
        return string.Format("({0},{1},{2})",
          RealString(p.X), RealString(p.Y),
          RealString(p.Z));
    }

    public static string RealString(double a)
    {
        return a.ToString("0.##");
    }
  }
}
