#region Copyright
//
// Copyright (C) 2010-2011 by Autodesk, Inc.
//
// Permission to use, copy, modify, and distribute this software in
// object code form for any purpose and without fee is hereby granted,
// provided that the above copyright notice appears in all copies and
// that both that copyright notice and the limited warranty and
// restricted rights notice below appear in all supporting
// documentation.
//
// AUTODESK PROVIDES THIS PROGRAM "AS IS" AND WITH ALL FAULTS.
// AUTODESK SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTY OF
// MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.  AUTODESK, INC.
// DOES NOT WARRANT THAT THE OPERATION OF THE PROGRAM WILL BE
// UNINTERRUPTED OR ERROR FREE.
//
// Use, duplication, or disclosure by the U.S. Government is subject to
// restrictions set forth in FAR 52.227-19 (Commercial Computer
// Software - Restricted Rights) and DFAR 252.227-7013(c)(1)(ii)
// (Rights in Technical Data and Computer Software), as applicable.
//
// 
#endregion // Copyright

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using Autodesk.Revit.Attributes;
using Autodesk.Revit.DB;
using Autodesk.Revit.DB.Mechanical;
using Autodesk.Revit.UI;

namespace Revit.Test.MepPlaceholders
{

  /// <summary>
  /// get element's id and type information and its supported information.
  /// </summary>
  [Transaction( TransactionMode.Manual )]
  public class MEPPlaceholders : IExternalCommand
  {
    class EquipmentElement
    {
      public FamilyInstance FamilyInstance;
      public Connector SupplyAirConnector;
      public XYZ ConnectionPoint;
      public XYZ ConnectionDirection;

      public EquipmentElement( 
        FamilyInstance familyInstance, 
        Connector supplyAirConnector, 
        XYZ connectionPoint, 
        XYZ connectionDirection )
      {
        FamilyInstance = familyInstance;
        SupplyAirConnector = supplyAirConnector;
        ConnectionPoint = connectionPoint;
        ConnectionDirection = connectionDirection;
      }
    }

    /// <summary>
    /// Implement this method as an external command for Revit.
    /// </summary>
    /// <param name="revit">An object that is passed to the external application 
    /// which contains data related to the command, 
    /// such as the application object and active view.</param>
    /// <param name="message">A message that can be set by the external application 
    /// which will be displayed if a failure or cancellation is returned by 
    /// the external command.</param>
    /// <param name="elements">A set of elements to which the external application 
    /// can add elements that are to be highlighted in case of failure or cancellation.</param>
    /// <returns>Return the status of the external command. 
    /// A result of Succeeded means that the API external method functioned as expected. 
    /// Cancelled can be used to signify that the user cancelled the external operation 
    /// at some point. Failure should be returned if the application is unable to proceed with 
    /// the operation.</returns>
    public Result Execute( 
      ExternalCommandData revit,
      ref string message,
      ElementSet elements )
    {
      m_uiApp = revit.Application;
      m_app = m_uiApp.Application;
      m_uiDoc = m_uiApp.ActiveUIDocument;
      m_doc = m_uiDoc.Document;

      ExecuteMEPPlaceholders( m_doc );

      return Result.Succeeded;
    }

    private void ExecuteMEPPlaceholders( Document doc )
    {
      // Find duct type
      FilteredElementCollector ductTypeCollector = new FilteredElementCollector( doc );
      ductTypeCollector.OfClass( typeof( DuctType ) );
      Func<DuctType, bool> isRectangularRadiusDuctType = dt => dt.Name.Contains( "Radius Elbows / Tees" );
      DuctType ductType = ductTypeCollector.OfType<DuctType>().First<DuctType>( isRectangularRadiusDuctType );

      // Find level 1
      FilteredElementCollector levelCollector = new FilteredElementCollector( doc );
      levelCollector.OfClass( typeof( Level ) );
      Func<Level, bool> isLevel1 = level => level.Name == "Level 1";
      Level level1 = levelCollector.Cast<Level>().First<Level>( isLevel1 );


      // Find all mechanical equipment
      FilteredElementCollector collector = new FilteredElementCollector( doc );
      collector.OfClass( typeof( FamilyInstance ) );
      List<BuiltInCategory> categories = new List<BuiltInCategory>();
      categories.Add( BuiltInCategory.OST_MechanicalEquipment );
      categories.Add( BuiltInCategory.OST_DuctTerminal );
      collector.WherePasses( new ElementMulticategoryFilter( categories ) );

      IList<Element> elements = collector.ToElements();

      List<EquipmentElement> equipmentElements = new List<EquipmentElement>();


      foreach( FamilyInstance instance in elements.Cast<FamilyInstance>() )
      {

        // Find connectors
        ConnectorManager cm = instance.MEPModel.ConnectorManager;
        ConnectorSet cs = cm.Connectors;

        foreach( Connector connector in cs )
        {
          if( connector.Domain == Domain.DomainHvac && connector.DuctSystemType == DuctSystemType.SupplyAir )
          {
            XYZ connectorPoint = connector.Origin;
            XYZ connectorDirection = connector.CoordinateSystem.BasisZ;

            equipmentElements.Add( new EquipmentElement( instance, connector, connectorPoint, connectorDirection ) );
          }
        }
      }

      // Find a waypoint for the cross fitting
      double verticalLocation = 11 + level1.Elevation;


      double[] min = new double[] { double.PositiveInfinity, double.PositiveInfinity };
      double[] max = new double[] { double.NegativeInfinity, double.NegativeInfinity };
      foreach( EquipmentElement e in equipmentElements )
      {
        BoundingBoxXYZ box = e.FamilyInstance.get_BoundingBox( null );
        for( int i = 0; i < 2; i++ )
        {
          min[i] = box.Min[i] < min[i] ? box.Min[i] : min[i];
          max[i] = box.Max[i] > max[i] ? box.Max[i] : max[i];
        }
      }

      XYZ wayPoint = new XYZ( ( min[0] + max[0] ) / 2, ( min[1] + max[1] ) / 2, verticalLocation );


      TransactionGroup tGroup = new TransactionGroup( doc );
      tGroup.Start( "Auto-route placeholders" );

      // Create system

      Transaction tSystem = new Transaction( doc, "Create system" );
      tSystem.Start();
      Connector baseConnector = null;
      ConnectorSet newSystemCS = new ConnectorSet();
      foreach( EquipmentElement e in equipmentElements )
      {
        if( e.FamilyInstance.Category.Id == new ElementId( BuiltInCategory.OST_MechanicalEquipment ) )
          baseConnector = e.SupplyAirConnector;
        else
          newSystemCS.Insert( e.SupplyAirConnector );
      }
      doc.Create.NewMechanicalSystem( baseConnector, newSystemCS, DuctSystemType.SupplyAir );
      tSystem.Commit();

      bool xFirst = true;
      List<Connector> wayPointConnectors = new List<Connector>();
      foreach( EquipmentElement e in equipmentElements )
      {
        Connector nextConnector;
        //if connector direction is vertical, add duct to reach target elevation
        if( !e.ConnectionDirection.IsAlmostEqualTo( XYZ.BasisZ ) )
        {
          throw new NotImplementedException( "Not implemented for initially non-vertical connectors" );
        }

        Transaction t = new Transaction( doc, "Create duct" );
        t.Start();
        Duct duct = Duct.CreatePlaceholder( doc, ductType.Id, level1.Id, e.ConnectionPoint, new XYZ( e.ConnectionPoint.X, e.ConnectionPoint.Y, wayPoint.Z ) );
        t.Commit();
        t.Start( "Connect duct" );
        Connector targetConnector = GetDuctConnectorAt( duct, e.ConnectionPoint, out nextConnector );
        targetConnector.ConnectTo( e.SupplyAirConnector );
        t.Commit();

        // All connections should make a right hand turn into the waypoint
        XYZ nextConnectorPoint = nextConnector.Origin;
        XYZ nextWayPoint = null;
        if( xFirst )
        {
          nextWayPoint = new XYZ( wayPoint.X, nextConnectorPoint.Y, wayPoint.Z );
        }
        else
        {
          nextWayPoint = new XYZ( nextConnectorPoint.X, wayPoint.Y, wayPoint.Z );
        }
        t.Start( "Create duct" );
        Duct nextDuct = Duct.CreatePlaceholder( doc, ductType.Id, level1.Id, nextConnectorPoint, nextWayPoint );
        t.Commit();
        t.Start( "Add fitting" );
        Connector nextNextConnector;
        Connector nextTargetConnector = GetDuctConnectorAt( nextDuct, nextConnectorPoint, out nextNextConnector );
        doc.Create.NewElbowFitting( nextConnector, nextTargetConnector );
        t.Commit();

        t.Start( "Create duct" );
        nextDuct = Duct.CreatePlaceholder( doc, ductType.Id, level1.Id, nextNextConnector.Origin, wayPoint );
        t.Commit();
        t.Start( "Add fitting" );
        Connector lastConnector;
        Connector nextNextTargetConnector = GetDuctConnectorAt( nextDuct, nextNextConnector.Origin, out lastConnector );
        doc.Create.NewElbowFitting( nextNextConnector, nextNextTargetConnector );
        wayPointConnectors.Add( lastConnector );
        t.Commit();

        xFirst = !xFirst;
      }

      if( wayPointConnectors.Count != 4 )
        throw new Exception( "Unexpected number of connectors" );
      Transaction tCross = new Transaction( doc, "Add cross fitting" );
      tCross.Start();
      doc.Create.NewCrossFitting( wayPointConnectors[0], wayPointConnectors[2], wayPointConnectors[1], wayPointConnectors[3] );
      tCross.Commit();

      tGroup.Assimilate();

    }

    private Connector GetDuctConnectorAt( 
      Duct duct, 
      XYZ location, 
      out Connector otherConnector )
    {
      Connector targetConnector = null;
      otherConnector = null;
      ConnectorManager cm = duct.ConnectorManager;
      foreach( Connector connector in cm.Connectors )
      {
        if( connector.Origin.IsAlmostEqualTo( location ) )
        {
          targetConnector = connector;
        }
        else
        {
          otherConnector = connector;
        }
      }
      return targetConnector;
    }

    private Autodesk.Revit.UI.UIApplication m_uiApp;
    private Autodesk.Revit.UI.UIDocument m_uiDoc;
    private Autodesk.Revit.ApplicationServices.Application m_app;
    private Autodesk.Revit.DB.Document m_doc;

  }


  [Transaction( TransactionMode.Manual )]
  public class ConvertPlaceholders : IExternalCommand
  {
    public Result Execute( 
      ExternalCommandData commandData, 
      ref string message, 
      ElementSet elements )
    {
      Document doc = commandData.View.Document;

      Transaction t = new Transaction( doc, "Convert placeholder network" );
      t.Start();

      FilteredElementCollector ductCollector = new FilteredElementCollector( doc );
      ductCollector.OfClass( typeof( Duct ) );
      Func<Duct, bool> isPlaceholder = duct => duct.IsPlaceholder;

      IEnumerable<Duct> ducts = ductCollector.OfType<Duct>().Where<Duct>( isPlaceholder );

      ICollection<ElementId> ductIds = ducts.Select<Duct, ElementId>( duct => duct.Id ).ToList<ElementId>();

      MechanicalUtils.ConvertDuctPlaceholders( doc, ductIds );

      t.Commit();

      return Result.Succeeded;
    }
  }
}
